#!/usr/bin/python3

import os
import sys
import subprocess
import json
import locale
import tempfile
import zipfile
import shutil
import datetime
import copy
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
from PIL import Image
import requests

import gi
gi.require_version('Gio', '2.0')
from gi.repository import Gio, GLib

from . import logger

DEBUG = os.getenv("DEBUG") is not None

def debug(msg):
    if DEBUG:
        print(msg, file=sys.stderr)

def warn(msg):
    print(msg, file=sys.stderr)


class AbortedError(Exception):
    """Raised by a progress_callback to interrupt an in-flight harvester download."""

LANGUAGE_CODE = "C"
try:
    LANGUAGE_CODE = locale.getlocale()[0].split("_")[0]
except:
    pass

home = os.path.expanduser("~")
user_data_dir = GLib.get_user_data_dir()
locale_inst = os.path.join(user_data_dir, 'locale')
settings_dir = os.path.join(GLib.get_user_config_dir(), 'cinnamon', 'spices')
old_settings_dir = f'{home}/.cinnamon/configs/'

URL_SPICES_HOME = "https://cinnamon-spices.linuxmint.com"

SPICE_MAP = {
    "applet": {
        "url": URL_SPICES_HOME + "/json/applets.json",
        "enabled-schema": "org.cinnamon",
        "enabled-key": "enabled-applets",
        "install-folders": (os.path.join(user_data_dir, "cinnamon/applets/"),)
    },
    "desklet": {
        "url": URL_SPICES_HOME + "/json/desklets.json",
        "enabled-schema": "org.cinnamon",
        "enabled-key": "enabled-desklets",
        "install-folders": (os.path.join(user_data_dir, "cinnamon/desklets/"),)
    },
    "extension": {
        "url": URL_SPICES_HOME + "/json/extensions.json",
        "enabled-schema": "org.cinnamon",
        "enabled-key": "enabled-extensions",
        "install-folders": (os.path.join(user_data_dir, "cinnamon/extensions/"),)
    },
    "search_provider": {
        "url": "",
        "enabled-schema": "org.cinnamon",
        "enabled-key": "enabled-search-providers",
        "install-folders": (os.path.join(user_data_dir, "cinnamon/search_providers/"),)
    },
    "action": {
        "url": URL_SPICES_HOME + "/json/actions.json",
        "enabled-schema": "org.nemo.plugins",
        "enabled-key": "disabled-actions",
        "install-folders": (os.path.join(user_data_dir, "nemo/actions/"),)
    },
    "theme": {
        "url": URL_SPICES_HOME + "/json/themes.json",
        "enabled-schema": "org.cinnamon.theme",
        "enabled-key": "name",
        "install-folders": (
            os.path.join(home, ".themes/"),
            os.path.join(user_data_dir, "themes/"),
            os.path.join(user_data_dir, "cinnamon/themes/"),
        )
    }
}

TIMEOUT_DOWNLOAD_JSON = 15
TIMEOUT_DOWNLOAD_THUMB = 60
TIMEOUT_DOWNLOAD_ZIP = 120


def remove_empty_folders(path):
    if not os.path.isdir(path):
        return

    for entry in os.listdir(path):
        full = os.path.join(path, entry)
        if os.path.isdir(full):
            remove_empty_folders(full)

    if not os.listdir(path):
        debug(f"Removing empty folder: {path}")
        try:
            os.rmdir(path)
        except OSError:
            pass

activity_logger = logger.ActivityLogger()

# return how many times 10m goes into the utc timestamp.
# This gives us a unique value every 10 minutes to allow
# the server cache to be utilized.
TIMESTAMP_LIFETIME_MINUTES = 10


def get_current_timestamp():
    seconds = datetime.datetime.utcnow().timestamp()
    return int(seconds // (TIMESTAMP_LIFETIME_MINUTES * 60))

class SpiceUpdate:
    def __init__(self, spice_type, uuid, index_node, meta_node):

        self.uuid = uuid
        self.spice_type = spice_type

        self.author = ""
        try:
            author = index_node["author_user"]
            if author not in {"none", "unknown"}:
                self.author = author
        except:
            pass

        try:
            self.name = index_node["translations"][f"name_{LANGUAGE_CODE}"]
        except:
            self.name = index_node["name"]

        try:
            self.description = index_node["translations"][f"description_{LANGUAGE_CODE}"]
        except:
            self.description = index_node["description"]

        try:
            self.old_version = datetime.datetime.fromtimestamp(meta_node["last-edited"]).strftime("%Y.%m.%d")
        except:
            self.old_version = ""

        try:
            self.new_version = datetime.datetime.fromtimestamp(index_node["last_edited"]).strftime("%Y.%m.%d")
        except:
            self.new_version = ""

        self.commit_id = index_node['last_commit']
        self.commit_msg = index_node['last_commit_subject']

        self.link = f"{URL_SPICES_HOME}/{spice_type}s/view/{index_node['spices-id']}"
        self.size = index_node['file_size']


class SpicePathSet:
    def __init__(self, cache_item, spice_type):
        cache_folder = Path(os.path.join(GLib.get_user_cache_dir(), 'cinnamon', 'spices', spice_type))

        is_theme = spice_type == "theme"

        if is_theme:
            thumb_rel_path = Path(cache_item["screenshot"])
            self.thumb_basename = thumb_rel_path.name
            self.thumb_download_url = URL_SPICES_HOME + thumb_rel_path.as_posix()
        else:
            thumb_rel_path = Path(cache_item["icon"])
            self.thumb_basename = thumb_rel_path.name
            self.thumb_download_url = URL_SPICES_HOME + thumb_rel_path.as_posix()

        self.thumb_local_path = cache_folder.joinpath(self.thumb_basename)
        self.zip_download_url = URL_SPICES_HOME + cache_item['file']


class Harvester:
    def __init__(self, spice_type):
        self.spice_type = spice_type

        self.themes = self.spice_type == "theme"
        self.actions = self.spice_type == "action"

        self.has_cache = False
        self.updates = []
        self.meta_map = {}
        self.index_cache = {}
        self._previous_index = {}

        self.cache_folder = os.path.join(GLib.get_user_cache_dir(), 'cinnamon', 'spices', self.spice_type)

        self.index_file = os.path.join(self.cache_folder, "index.json")

        # User-writable install destinations from SPICE_MAP. Index 0 is the
        # canonical location for new installs; later entries are alternative
        # paths a spice may already live in (e.g. legacy theme dirs).
        self.install_folders = SPICE_MAP[self.spice_type]["install-folders"]
        self.install_folder = self.install_folders[0]

        # Full scan list = install_folders + read-only system locations. The
        # system entries surface system-installed spices in meta_map for the
        # cs_settings UI listing but are never written to.
        if self.actions:
            sys_dirs = tuple(os.path.join(d, 'nemo/actions/') for d in GLib.get_system_data_dirs())
        elif self.themes:
            sys_dirs = ()
        else:
            sys_dirs = (f'/usr/share/cinnamon/{self.spice_type}s/',)
        self.spices_directories = self.install_folders + sys_dirs

        self.settings = Gio.Settings.new(SPICE_MAP[self.spice_type]["enabled-schema"])
        self.enabled_key = SPICE_MAP[self.spice_type]["enabled-key"]

        try:
            self._load_cache()
        except ValueError:
            # Corrupt on-disk cache: file has been removed by _load_cache.
            # A subsequent refresh() will repopulate and surface any error.
            pass
        self._load_metadata()

    def anything_installed(self):
        for location in self.install_folders:
            path = Path(location)
            if not path.exists():
                continue

            for _ in path.iterdir():
                return True

        debug(f"No additional {self.spice_type}s installed")
        return False

    def refresh(self, full, force=False, progress_callback=None):
        if not force and not self.anything_installed():
            return

        debug(f"Cache stamp: {get_current_timestamp()}")

        os.makedirs(self.cache_folder, mode=0o755, exist_ok=True)
        self._update_local_json()

        if full:
            self._update_local_thumbs(progress_callback=progress_callback)

        self._load_metadata()
        self._clean_old_thumbs()

    def reload(self):
        self._load_metadata()
        return self._generate_update_list()

    def get_updates(self):
        return self._generate_update_list()

    def has_update(self, uuid):
        return self._spice_has_update(uuid)

    def install(self, uuid, progress_callback=None):
        self._install_by_uuid(uuid, progress_callback=progress_callback)

    def install_from_folder(self, folder, uuid, from_spices=False):
        """Install a spice from a local folder. Used by developer tools that
        bypass the spices server (e.g. cinnamon-install-spice)."""
        self._install_from_folder(folder, os.path.dirname(folder), uuid, from_spices=from_spices)
        self._load_metadata()

    def uninstall(self, uuid):
        self._uninstall_by_uuid(uuid)

    def get_enabled(self, uuid):
        if self.themes:
            return 1 if self.settings.get_string(self.enabled_key) == uuid else 0

        if self.actions:
            disabled_list = self.settings.get_strv(self.enabled_key)
            return 0 if f"{uuid}.nemo_action" in disabled_list else 1

        enabled_count = 0
        for item in self.settings.get_strv(self.enabled_key):
            item = item.replace("!", "")
            if uuid in item.split(":"):
                enabled_count += 1
        return enabled_count

    def enable(self, uuid, panel=1, box='right', position=0, desklet_x=100, desklet_y=100):
        if self.spice_type == 'applet':
            applet_id = self.settings.get_int('next-applet-id')
            self.settings.set_int('next-applet-id', applet_id + 1)

            entries = []
            for entry in self.settings.get_strv(self.enabled_key):
                info = entry.split(':')
                pos = int(info[2])
                if info[0] == f'panel{panel}' and info[1] == box and position <= pos:
                    info[2] = str(pos + 1)
                    entries.append(':'.join(info))
                else:
                    entries.append(entry)
            entries.append(f'panel{panel}:{box}:{position}:{uuid}:{applet_id}')
            self.settings.set_strv(self.enabled_key, entries)
        elif self.spice_type == 'desklet':
            desklet_id = self.settings.get_int('next-desklet-id')
            self.settings.set_int('next-desklet-id', desklet_id + 1)
            enabled = self.settings.get_strv(self.enabled_key)
            enabled.append(f'{uuid}:{desklet_id}:{desklet_x}:{desklet_y}')
            self.settings.set_strv(self.enabled_key, enabled)
        elif self.actions:
            disabled = self.settings.get_strv(self.enabled_key)
            uuid_name = f"{uuid}.nemo_action"
            if uuid_name in disabled:
                disabled.remove(uuid_name)
                self.settings.set_strv(self.enabled_key, disabled)
        elif not self.themes:
            enabled = self.settings.get_strv(self.enabled_key)
            enabled.append(uuid)
            self.settings.set_strv(self.enabled_key, enabled)

    def disable(self, uuid):
        if self.themes:
            return

        if self.actions:
            disabled = self.settings.get_strv(self.enabled_key)
            uuid_name = f"{uuid}.nemo_action"
            if uuid_name not in disabled:
                disabled.append(uuid_name)
                self.settings.set_strv(self.enabled_key, disabled)
            return

        new_list = []
        for entry in self.settings.get_strv(self.enabled_key):
            if self.spice_type == 'applet':
                entry_uuid = entry.split(':')[3].lstrip('!')
            elif self.spice_type == 'desklet':
                entry_uuid = entry.split(':')[0].lstrip('!')
            else:
                entry_uuid = entry.lstrip('!')
            if entry_uuid != uuid:
                new_list.append(entry)
        self.settings.set_strv(self.enabled_key, new_list)

    def _update_local_json(self):
        debug(f"harvester: Downloading new list of available {self.spice_type}s")
        url = SPICE_MAP[self.spice_type]["url"]

        try:
            r = requests.get(url,
                             timeout=TIMEOUT_DOWNLOAD_JSON,
                             params={"time": get_current_timestamp()})
            debug(f"Downloading from {r.request.url}")
            r.raise_for_status()
        except Exception as e:
            warn(f"Could not refresh json data for {self.spice_type}: {e}")
            raise

        # Validate before overwriting the on-disk cache so a malformed
        # response doesn't clobber the previously-good index.
        try:
            new_cache = json.loads(r.text)
        except ValueError as e:
            warn(f"Server returned malformed JSON for {self.spice_type}: {e}")
            raise

        with open(self.index_file, "w", encoding="utf-8") as f:
            f.write(r.text)

        # Snapshot the prior index so _download_thumb can tell whether the
        # server-side last_edited moved for each uuid.
        self._previous_index = self.index_cache
        self.index_cache = new_cache
        self.has_cache = True

    def _update_local_thumbs(self, progress_callback=None):
        # This uses threads for the downloads, but this function blocks until
        # all are downloaded. Per-thumb progress is reported on the calling
        # thread as each future completes.
        items = list(self.index_cache.items())

        total = len(items)
        if total == 0:
            return

        tpe = ThreadPoolExecutor(max_workers=10)
        try:
            futures = []
            for uuid, item in items:
                debug(f"Submitting thumb_job for {uuid}")
                futures.append(tpe.submit(self._download_thumb, copy.copy(uuid), copy.deepcopy(item)))

            for done, future in enumerate(futures, start=1):
                future.result()
                if progress_callback is not None:
                    progress_callback(done, 1, total)
        finally:
            # If progress_callback raised AbortedError, drop still-queued thumbs.
            # In-flight ones (up to max_workers) finish on their own.
            tpe.shutdown(wait=False, cancel_futures=True)

    def _download_thumb(self, uuid, item):
        paths = SpicePathSet(item, spice_type=self.spice_type)
        prev = self._previous_index.get(uuid)
        server_changed = prev is None or prev.get("last_edited") != item.get("last_edited")
        needs_download = (not os.path.isfile(paths.thumb_local_path)
                          or self._is_bad_image(paths.thumb_local_path)
                          or server_changed)
        if not needs_download:
            return

        debug(f"Downloading thumbnail for {uuid}: {paths.thumb_download_url}")

        try:
            r = requests.get(paths.thumb_download_url,
                             timeout=TIMEOUT_DOWNLOAD_THUMB,
                             params={"time": get_current_timestamp()})
            r.raise_for_status()
        except (requests.RequestException, OSError) as e:
            warn(f"Could not get thumbnail for {uuid}: {e}")
            return

        with open(paths.thumb_local_path, "wb") as f:
            f.write(r.content)

    def _load_metadata(self):
        debug(f"harvester: Loading metadata on installed {self.spice_type}s")

        new_map = {}

        for directory in self.spices_directories:
            if not os.path.isdir(directory):
                continue

            for entry in os.listdir(directory):
                full_path = os.path.join(directory, entry)

                if self.actions:
                    if entry == 'sample.nemo_action':
                        continue
                    if entry.endswith('.nemo_action'):
                        sibling_dir = full_path[:-len('.nemo_action')]
                        if os.path.isdir(sibling_dir):
                            # Spice-installed action: handled via the sibling dir below.
                            continue
                        uuid = entry[:-len('.nemo_action')]
                        if uuid in new_map:
                            continue
                        self._load_keyfile_action(full_path, entry, new_map)
                        continue
                    if not os.path.isdir(full_path):
                        # Other top-level files (helper scripts, etc.)
                        continue

                if not os.path.isdir(full_path):
                    continue

                # First-wins: spices_directories iterates user install_folders
                # before system paths, so the user-installed copy of a uuid
                # masks any system copy and a just-installed upgrade isn't
                # shadowed by an orphan in another user-level path.
                if entry in new_map:
                    continue

                try:
                    with open(os.path.join(full_path, "metadata.json"), "r", encoding="utf-8") as f:
                        metadata = json.load(f)
                        metadata['path'] = full_path
                        metadata['writable'] = os.access(full_path, os.W_OK)
                        new_map[entry] = metadata
                except FileNotFoundError:
                    if not self.themes:
                        warn(f"Skipping {entry}: no metadata.json")
                except Exception as detail:
                    warn(detail)
                    warn(f"Skipping {entry}: there was a problem trying to read metadata.json")

        self.meta_map = new_map

    def _load_keyfile_action(self, full_path, entry, target_map):
        metadata = {}
        keyfile = GLib.KeyFile.new()

        try:
            keyfile.load_from_file(full_path, GLib.KeyFileFlags.KEEP_TRANSLATIONS)
        except GLib.Error as e:
            warn(f"Could not read action file '{full_path}': {e.message}")
            return

        try:
            # Skip actions explicitly marked Active=false. A missing key is
            # treated as active; any other read failure is treated as opt-out.
            if not keyfile.get_boolean('Nemo Action', 'Active'):
                return
        except GLib.Error as e:
            if e.code != GLib.KeyFileError.KEY_NOT_FOUND:
                warn(f"Could not read Active field for action '{full_path}': {e.message}")
                return

        try:
            name = keyfile.get_locale_string('Nemo Action', 'Name')
            metadata['name'] = name.replace("_", "")
        except GLib.Error as e:
            warn(f"Could not read Name field for action. Skipping '{full_path}': {e.message}")
            return

        try:
            metadata['description'] = keyfile.get_locale_string('Nemo Action', 'Comment')
        except GLib.Error as e:
            if e.code != GLib.KeyFileError.KEY_NOT_FOUND:
                warn(f"Could not read Comment field for action '{full_path}': {e.message}")
            else:
                metadata['description'] = ""

        try:
            metadata['icon'] = keyfile.get_string('Nemo Action', 'Icon-Name')
        except GLib.Error as e:
            if e.code != GLib.KeyFileError.KEY_NOT_FOUND:
                warn(f"Could not read Icon-Name field for action '{full_path}': {e.message}")

        uuid = entry[:-len('.nemo_action')]
        # ManageSpicesRow expects a directory it can scan for version subdirs.
        metadata['path'] = os.path.dirname(full_path)
        metadata['writable'] = False
        metadata['disable_about'] = True
        metadata['uuid'] = uuid
        target_map[uuid] = metadata

    def _load_cache(self):
        debug(f"harvester: Loading local {self.spice_type} cache")

        try:
            with open(self.index_file, "r", encoding="utf-8") as f:
                self.index_cache = json.load(f)
                self.has_cache = True
        except FileNotFoundError:
            self.index_cache = {}
            self.has_cache = False
            return
        except ValueError as e:
            try:
                os.remove(self.index_file)
            except:
                pass
            self.index_cache = {}
            self.has_cache = False
            raise

    def _generate_update_list(self):
        debug(f"harvester: Generating list of updated {self.spice_type}s")

        self.updates = []

        for uuid in self.index_cache:
            try:
                if uuid in self.meta_map and self._spice_has_update(uuid):
                    update = SpiceUpdate(self.spice_type, uuid, self.index_cache[uuid], self.meta_map[uuid])
                    self.updates.append(update)
            except Exception as e:
                warn(f"Error checking updates for {uuid}: {e}")

        return self.updates

    def _spice_has_update(self, uuid):
        # System-installed xlets ship without a `last-edited` field, so the
        # int() lookup below raises and we naturally answer False. Spice-installed
        # copies (in any user dir) get the timestamp written during install.
        try:
            return int(self.meta_map[uuid]["last-edited"]) < self.index_cache[uuid]["last_edited"]
        except Exception:
            return False

    def _install_by_uuid(self, uuid, progress_callback=None):
        action = "upgrade" if uuid in self.meta_map else "install"

        try:
            item = self.index_cache[uuid]
        except KeyError:
            warn(f"Can't install {uuid} - it doesn't seem to exist on the server")
            raise

        paths = SpicePathSet(item, spice_type=self.spice_type)

        tmp_name = None
        try:
            with tempfile.NamedTemporaryFile(delete=False) as tmp:
                tmp_name = tmp.name
                self._download_zip(paths.zip_download_url, tmp, progress_callback)

            with zipfile.ZipFile(tmp_name) as _zip:
                with tempfile.TemporaryDirectory() as d:
                    for member in _zip.infolist():
                        _zip.extract(member, d)
                        os.chmod(os.path.join(d, member.filename), member.external_attr >> 16)
                    self._install_from_folder(os.path.join(d, uuid), d, uuid, from_spices=True)
                    self.write_to_log(uuid, action)

            if self.actions and action == "install":
                # Newly installed actions start in the disabled list so the user
                # has to opt in to running them.
                disabled = self.settings.get_strv(self.enabled_key)
                uuid_name = f"{uuid}.nemo_action"
                if uuid_name not in disabled:
                    disabled.append(uuid_name)
                    self.settings.set_strv(self.enabled_key, disabled)
        except AbortedError:
            raise
        except Exception as e:
            warn(f"couldn't install: {e}")
            raise
        finally:
            if tmp_name is not None:
                try:
                    os.remove(tmp_name)
                except FileNotFoundError:
                    pass
            self._load_metadata()

    def _download_zip(self, url, out_file, progress_callback):
        block_size = 16 * 1024

        try:
            r = requests.get(url,
                             timeout=TIMEOUT_DOWNLOAD_ZIP,
                             params={"time": get_current_timestamp()},
                             stream=True)
            r.raise_for_status()
        except Exception as e:
            warn(f"Could not start zip download: {e}")
            raise

        try:
            try:
                total_size = int(r.headers.get("content-length", 0))
            except (TypeError, ValueError):
                total_size = 0

            count = 0
            for chunk in r.iter_content(chunk_size=block_size):
                if not chunk:
                    continue
                out_file.write(chunk)
                count += 1
                if progress_callback is not None:
                    progress_callback(count, block_size, total_size)
        finally:
            r.close()

    def _uninstall_by_uuid(self, uuid):
        try:
            if not self.themes:
                if os.path.exists(locale_inst):
                    for i19_folder in os.listdir(locale_inst):
                        mo_path = os.path.join(locale_inst, i19_folder, 'LC_MESSAGES', f'{uuid}.mo')
                        if os.path.isfile(mo_path):
                            os.remove(mo_path)
                        remove_empty_folders(os.path.join(locale_inst, i19_folder))

                for cfg in (settings_dir, old_settings_dir):
                    cfg_path = os.path.join(cfg, uuid)
                    if os.path.exists(cfg_path):
                        shutil.rmtree(cfg_path)

            self._remove_spice_from_all_directories(uuid)

            if self.actions:
                disabled_list = self.settings.get_strv(self.enabled_key)
                uuid_name = f"{uuid}.nemo_action"
                if uuid_name in disabled_list:
                    disabled_list.remove(uuid_name)
                    self.settings.set_strv(self.enabled_key, disabled_list)

            self.write_to_log(uuid, "remove")
        except Exception as e:
            warn(f"couldn't uninstall {uuid}: {e}")
            raise
        finally:
            self._load_metadata()

    def _install_from_folder(self, folder, base_folder, uuid, from_spices=False):
        contents = os.listdir(folder)

        if not self.themes:
            # Install spice localization files, if any
            if 'po' in contents:
                po_dir = os.path.join(folder, 'po')
                for file in os.listdir(po_dir):
                    if file.endswith('.po'):
                        lang = file.split(".")[0]
                        locale_dir = os.path.join(locale_inst, lang, 'LC_MESSAGES')
                        os.makedirs(locale_dir, mode=0o755, exist_ok=True)
                        subprocess.run(['/usr/bin/msgfmt', '-c',
                                       os.path.join(po_dir, file), '-o',
                                       os.path.join(locale_dir, f'{uuid}.mo')],
                                       check=True)

        os.makedirs(self.install_folder, mode=0o755, exist_ok=True)

        # Clear prior copies from every user-writable location, not just
        # install_folder, so an upgrade across a path change doesn't leave
        # an orphan that the first-wins scan would surface as the active copy.
        self._remove_spice_from_all_directories(uuid)

        dest = os.path.join(self.install_folder, uuid)
        if not self.actions:
            shutil.copytree(folder, dest)
        else:
            shutil.copytree(base_folder, self.install_folder, dirs_exist_ok=True)

        meta_path = os.path.join(dest, 'metadata.json')

        if self.themes and not os.path.exists(meta_path):
            md = {}
        else:
            with open(meta_path, "r", encoding='utf-8') as f:
                md = json.load(f)

        if from_spices and uuid in self.index_cache:
            md['last-edited'] = self.index_cache[uuid]['last_edited']
        else:
            md['last-edited'] = int(datetime.datetime.utcnow().timestamp())

        with open(meta_path, "w+", encoding='utf-8') as f:
            json.dump(md, f, indent=4)

    def _remove_spice_from_all_directories(self, uuid):
        # Iterates user-writable install_folders only — system spices are
        # package-managed and not ours to delete from /usr/share/.
        for directory in self.install_folders:
            dest = os.path.join(directory, uuid)
            if os.path.isdir(dest):
                shutil.rmtree(dest, ignore_errors=True)
            if self.actions:
                action_file = os.path.join(directory, f"{uuid}.nemo_action")
                try:
                    os.remove(action_file)
                except FileNotFoundError:
                    pass

    def write_to_log(self, uuid, action):
        new_version = "<none>"
        old_version = "<none>"

        try:
            remote_item = self.index_cache[uuid]
            new_version = datetime.datetime.fromtimestamp(remote_item["last_edited"]).strftime("%Y.%m.%d")
        except KeyError:
            if action in ("upgrade", "install"):
                warn(f"Upgrading {uuid} with no local metadata - something's wrong")

        try:
            local_item = self.meta_map[uuid]
            old_version = datetime.datetime.fromtimestamp(local_item["last-edited"]).strftime("%Y.%m.%d")
        except KeyError:
            if action in ("upgrade", "remove"):
                warn(f"Upgrading or removing {uuid} with no local metadata - something's wrong")

        log_timestamp = datetime.datetime.now().strftime("%F %T")
        activity_logger.log(f"{log_timestamp} {self.spice_type} {action} {uuid} {old_version} {new_version}")

    def _is_bad_image(self, path):
        try:
            Image.open(path)
        except IOError:
            return True
        return False

    def _clean_old_thumbs(self):
        # Cleanup obsolete thumbs
        flist = os.listdir(self.cache_folder)

        for item in self.index_cache.values():
            paths = SpicePathSet(item, spice_type=self.spice_type)
            try:
                flist.remove(paths.thumb_basename)
            except ValueError:
                pass

        for f in flist:
            if f == "index.json":
                continue
            try:
                debug(f"removing old thumb: {f}")
                os.remove(os.path.join(self.cache_folder, f))
            except:
                pass



#!/usr/bin/python3

from bin import util
util.strip_syspath_locals()

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('XApp', '1.0')

import os
import sys
from setproctitle import setproctitle
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), "bin"))
import gettext
import json
import argparse
import importlib.util
import traceback
from pathlib import Path

from bin.JsonSettingsWidgets import *
from bin.ExtensionCore import find_extension_subdir
from gi.repository import Gtk, Gio, XApp, GLib

# i18n
gettext.install("cinnamon", "/usr/share/locale")

home = os.path.expanduser("~")
settings_dir = os.path.join(GLib.get_user_config_dir(), 'cinnamon', 'spices')
old_settings_dir = '%s/.cinnamon/configs/' % home

translations = {}

proxy = None

XLET_SETTINGS_WIDGETS = {
    "entry"             :   "JSONSettingsEntry",
    "textview"          :   "JSONSettingsTextView",
    "checkbox"          :   "JSONSettingsSwitch", # deprecated: please use switch instead
    "switch"            :   "JSONSettingsSwitch",
    "spinbutton"        :   "JSONSettingsSpinButton",
    "filechooser"       :   "JSONSettingsFileChooser",
    "scale"             :   "JSONSettingsRange",
    "radiogroup"        :   "JSONSettingsComboBox", # deprecated: please use combobox instead
    "combobox"          :   "JSONSettingsComboBox",
    "colorchooser"      :   "JSONSettingsColorChooser",
    "fontchooser"       :   "JSONSettingsFontButton",
    "soundfilechooser"  :   "JSONSettingsSoundFileChooser",
    "iconfilechooser"   :   "JSONSettingsIconChooser",
    "datechooser"       :   "JSONSettingsDateChooser",
    "timechooser"       :   "JSONSettingsTimeChooser",
    "keybinding"        :   "JSONSettingsKeybinding",
    "list"              :   "JSONSettingsList"
}

class XLETSettingsButton(Button):
    def __init__(self, info, uuid, instance_id):
        super(XLETSettingsButton, self).__init__(info["description"])
        self.uuid = uuid
        self.instance_id = instance_id
        self.xletCallback = str(info["callback"])

    def on_activated(self):
        proxy.activateCallback('(sss)', self.xletCallback, self.uuid, self.instance_id)

def translate(uuid, string):
    #check for a translation for this xlet
    if uuid not in translations:
        try:
            translations[uuid] = gettext.translation(uuid, home + "/.local/share/locale").gettext
        except IOError:
            try:
                translations[uuid] = gettext.translation(uuid, "/usr/share/locale").gettext
            except IOError:
                translations[uuid] = None

    #do not translate whitespaces
    if not string.strip():
        return string

    if translations[uuid]:
        result = translations[uuid](string)

        try:
            result = result.decode("utf-8")
        except (AttributeError, UnicodeDecodeError):
            result = result

        if result != string:
            return result
    return _(string)

class MainWindow(object):
    def __init__(self, args):
        self.type = args.type
        self.uuid = args.uuid
        self.tab = 0
        self.instance_id = str(args.id)
        if args.tab is not None:
            self.tab = int(args.tab)

        self.selected_instance = None
        self.gsettings = Gio.Settings.new("org.cinnamon")
        self.custom_modules = {}

        self.load_xlet_data()
        self.build_window()
        self.load_instances()
        self.window.show_all()
        if self.instance_id and len(self.instance_info) > 1:
            for info in self.instance_info:
                if info["id"] == self.instance_id:
                    self.set_instance(info)
                    break
        else:
            self.set_instance(self.instance_info[0])
        try:
            Gio.DBusProxy.new_for_bus(Gio.BusType.SESSION, Gio.DBusProxyFlags.NONE, None,
                                      "org.Cinnamon", "/org/Cinnamon", "org.Cinnamon", None, self._on_proxy_ready, None)
        except dbus.exceptions.DBusException as e:
            print(e)

    def _on_proxy_ready (self, obj, result, data=None):
        global proxy
        proxy = Gio.DBusProxy.new_for_bus_finish(result)

        if not proxy.get_name_owner():
            proxy = None

        if proxy:
            proxy.highlightXlet('(ssb)', self.uuid, self.selected_instance["id"], True)

    def load_xlet_data (self):
        self.xlet_dir = f"/usr/share/cinnamon/{self.type}s/{self.uuid}"
        if not os.path.exists(self.xlet_dir):
            self.xlet_dir = f"{home}/.local/share/cinnamon/{self.type}s/{self.uuid}"

        if os.path.exists(f"{self.xlet_dir}/metadata.json"):
            with open(f"{self.xlet_dir}/metadata.json") as f:
                raw_data = f.read()
                self.xlet_meta = json.loads(raw_data)
        else:
            print(f"Could not find {self.type} metadata for uuid {self.uuid} - are you sure it's installed correctly?")
            quit()

    def build_window(self):
        self.window = XApp.GtkWindow()
        self.window.set_default_size(800, 600)
        main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.window.add(main_box)

        toolbar = Gtk.Toolbar()
        toolbar.get_style_context().add_class("primary-toolbar")
        toolbar.set_show_arrow(False)
        main_box.add(toolbar)

        toolitem = Gtk.ToolItem()
        toolitem.set_expand(True)
        toolbar.add(toolitem)
        toolbutton_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        toolitem.add(toolbutton_box)
        instance_button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        instance_button_box.get_style_context().add_class("linked")
        toolbutton_box.pack_start(instance_button_box, False, False, 0)

        self.prev_button = Gtk.Button.new_from_icon_name('xsi-go-previous-symbolic', Gtk.IconSize.BUTTON)
        self.prev_button.set_tooltip_text(_("Previous instance"))
        instance_button_box.add(self.prev_button)

        self.next_button = Gtk.Button.new_from_icon_name('xsi-go-next-symbolic', Gtk.IconSize.BUTTON)
        self.next_button.set_tooltip_text(_("Next instance"))
        instance_button_box.add(self.next_button)

        self.stack_switcher = Gtk.StackSwitcher()
        toolbutton_box.set_center_widget(self.stack_switcher)

        self.menu_button = Gtk.MenuButton()
        image = Gtk.Image.new_from_icon_name("xsi-open-menu-symbolic", Gtk.IconSize.BUTTON)
        self.menu_button.add(image)
        self.menu_button.set_tooltip_text(_("More options"))
        toolbutton_box.pack_end(self.menu_button, False, False, 0)

        menu = Gtk.Menu()
        menu.set_halign(Gtk.Align.END)

        restore_option = Gtk.MenuItem(label=_("Import from a file"))
        menu.append(restore_option)
        restore_option.connect("activate", self.restore)
        restore_option.show()

        backup_option = Gtk.MenuItem(label=_("Export to a file"))
        menu.append(backup_option)
        backup_option.connect("activate", self.backup)
        backup_option.show()

        reset_option = Gtk.MenuItem(label=_("Reset to defaults"))
        menu.append(reset_option)
        reset_option.connect("activate", self.reset)
        reset_option.show()

        separator = Gtk.SeparatorMenuItem()
        menu.append(separator)
        separator.show()

        reload_option = Gtk.MenuItem(label=_("Reload %s") % self.uuid)
        menu.append(reload_option)
        reload_option.connect("activate", self.reload_xlet)
        reload_option.show()

        self.menu_button.set_popup(menu)

        scw = Gtk.ScrolledWindow()
        scw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        main_box.pack_start(scw, True, True, 0)
        self.instance_stack = Gtk.Stack()
        scw.add(self.instance_stack)

        if "icon" in self.xlet_meta:
            self.window.set_icon_name(self.xlet_meta["icon"])

        icon_path = os.path.join(self.xlet_dir, "icon.svg")
        if os.path.exists(icon_path):
            self.window.set_icon_from_file(icon_path)
        else:
            icon_path = os.path.join(self.xlet_dir, "icon.png")
            if os.path.exists(icon_path):
                self.window.set_icon_from_file(icon_path)

        self.window.set_title(translate(self.uuid, self.xlet_meta["name"]))

        def check_sizing(widget, data=None):
            natreq = self.window.get_preferred_size()[1]
            monitor = Gdk.Display.get_default().get_monitor_at_window(self.window.get_window())

            height = monitor.get_workarea().height
            if natreq.height > height - 100:
                self.window.resize(800, 600)
                scw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

        self.window.connect("destroy", self.quit)
        self.window.connect("realize", check_sizing)
        self.prev_button.connect("clicked", self.previous_instance)
        self.next_button.connect("clicked", self.next_instance)

    def load_instances(self):
        self.instance_info = []
        path = Path(os.path.join(settings_dir, self.uuid))
        old_path = Path(f"{home}/.cinnamon/configs/{self.uuid}")
        instances = 0
        new_items = os.listdir(path) if path.exists() else []
        old_items = os.listdir(old_path) if old_path.exists() else []
        dir_items = sorted(new_items + old_items)
        try:
            multi_instance = int(self.xlet_meta["max-instances"]) != 1
        except (KeyError, ValueError):
            multi_instance = False

        for item in dir_items:
            # ignore anything that isn't json
            if item[-5:] != ".json":
                continue

            instance_id = item[0:-5]
            if not multi_instance and instance_id != self.uuid:
                continue # for single instance the file name should be [uuid].json

            if multi_instance:
                try:
                    int(instance_id)
                except (TypeError, ValueError):
                    traceback.print_exc()
                    continue # multi-instance should have file names of the form [instance-id].json

                instance_exists = False
                enabled = self.gsettings.get_strv(f'enabled-{self.type}s')
                for definition in enabled:
                    parts = definition.split(':')
                    if self.type == "applet" and len(parts) >= 5:
                        if parts[3] == self.uuid and parts[4] == instance_id:
                            instance_exists = True
                            break
                    elif self.type == "desklet" and len(parts) >= 2:
                        if parts[0] == self.uuid and parts[1] == instance_id:
                            instance_exists = True
                            break

                if not instance_exists:
                    continue

            settings = JSONSettingsHandler(
                os.path.join(path if item in new_items else old_path, item), self.uuid, instance_id, self.notify_dbus
            )
            instance_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            self.instance_stack.add_named(instance_box, instance_id)

            info = {"settings": settings, "id": instance_id}
            self.instance_info.append(info)

            settings_map = settings.get_settings()
            first_key = next(iter(settings_map.values()))

            try:
                for setting in settings_map:
                    if setting == "__md5__":
                        continue
                    for key in settings_map[setting]:
                        if key in ("description", "tooltip", "units"):
                            try:
                                settings_map[setting][key] = translate(self.uuid, settings_map[setting][key])
                            except (KeyError, ValueError):
                                traceback.print_exc()
                        elif key in "options":
                            new_opt_data = collections.OrderedDict()
                            opt_data = settings_map[setting][key]
                            for option in opt_data:
                                if opt_data[option] == "custom":
                                    continue
                                new_opt_data[translate(self.uuid, option)] = opt_data[option]
                            settings_map[setting][key] = new_opt_data
                        elif key in "columns":
                            columns_data = settings_map[setting][key]
                            for column in columns_data:
                                column["title"] = translate(self.uuid, column["title"])
            finally:
                # if a layout is not explicitly defined, generate the settings
                # widgets based on the order they occur
                if first_key["type"] == "layout":
                    self.build_with_layout(settings_map, info, instance_box, first_key)
                else:
                    self.build_from_order(settings_map, info, instance_box, first_key)

                if self.selected_instance is None:
                    self.selected_instance = info
                    if "stack" in info:
                        self.stack_switcher.set_stack(info["stack"])

            instances += 1

        if instances < 2:
            self.prev_button.set_no_show_all(True)
            self.next_button.set_no_show_all(True)

    def build_with_layout(self, settings_map, info, box, first_key):
        layout = first_key

        page_stack = SettingsStack()
        box.pack_start(page_stack, True, True, 0)
        self.stack_switcher.show()
        info["stack"] = page_stack

        for page_key in layout["pages"]:
            page_def = layout[page_key]
            if page_def['type'] == 'custom':
                page = self.create_custom_widget(page_def, info['settings'])
                if page is None:
                    continue
                elif not isinstance(page, SettingsPage):
                    print('page is not of type SettingsPage')
                    continue
            else:
                page = SettingsPage()
                for section_key in page_def["sections"]:
                    section_def = layout[section_key]
                    if 'dependency' in section_def:
                        revealer = JSONSettingsRevealer(info['settings'], section_def['dependency'])
                        section = page.add_reveal_section(translate(self.uuid, section_def["title"]), revealer=revealer)
                    else:
                        section = page.add_section(translate(self.uuid, section_def["title"]))
                    for key in section_def["keys"]:
                        item = settings_map[key]
                        settings_type = item["type"]
                        if settings_type == "button":
                            widget = XLETSettingsButton(item, self.uuid, info["id"])
                        elif settings_type == "label":
                            widget = Text(translate(self.uuid, item["description"]))
                        elif settings_type == 'custom':
                            widget = self.create_custom_widget(item, key, info['settings'])
                            if widget is None:
                                continue
                            elif not isinstance(widget, SettingsWidget):
                                print('widget is not of type SettingsWidget')
                                continue
                        elif settings_type in XLET_SETTINGS_WIDGETS:
                            widget = globals()[XLET_SETTINGS_WIDGETS[settings_type]](key, info["settings"], item)
                        else:
                            continue

                        if 'dependency' in item:
                            revealer = JSONSettingsRevealer(info['settings'], item['dependency'])
                            section.add_reveal_row(widget, revealer=revealer)
                        else:
                            section.add_row(widget)
            page_stack.add_titled(page, page_key, translate(self.uuid, page_def["title"]))

    def build_from_order(self, settings_map, info, box, first_key):
        page = SettingsPage()
        box.pack_start(page, True, True, 0)

        # if the first key is not of type 'header' or type 'section' we need to make a new section
        if first_key["type"] not in ("header", "section"):
            section = page.add_section(_("Settings for %s") % self.uuid)

        for key, item in settings_map.items():
            if key == "__md5__":
                continue
            if "type" in item:
                settings_type = item["type"]
                if settings_type in ("header", "section"):
                    if 'dependency' in item:
                        revealer = JSONSettingsRevealer(info['settings'], item['dependency'])
                        section = page.add_reveal_section(translate(self.uuid, item["description"]), revealer=revealer)
                    else:
                        section = page.add_section(translate(self.uuid, item["description"]))
                    continue

                if settings_type == "button":
                    widget = XLETSettingsButton(item, self.uuid, info["id"])
                elif settings_type == "label":
                    widget = Text(translate(self.uuid, item["description"]))
                elif settings_type == 'custom':
                    widget = self.create_custom_widget(item, key, info['settings'])
                    if widget is None:
                        continue
                    elif not isinstance(widget, SettingsWidget):
                        print('widget is not of type SettingsWidget')
                        continue
                elif settings_type in XLET_SETTINGS_WIDGETS:
                    widget = globals()[XLET_SETTINGS_WIDGETS[settings_type]](key, info["settings"], item)
                else:
                    continue

                if 'dependency' in item:
                    revealer = JSONSettingsRevealer(info['settings'], item['dependency'])
                    section.add_reveal_row(widget, revealer=revealer)
                else:
                    section.add_row(widget)

    def create_custom_widget(self, info, *args):
        file_name = info['file']
        widget_name = info['widget']
        file_path = os.path.join(find_extension_subdir(self.xlet_dir), file_name)

        try:
            if file_name not in self.custom_modules:
                spec = importlib.util.spec_from_file_location(self.uuid.replace('@', '') + '.' + file_name.split('.')[0], file_path)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                self.custom_modules[file_name] = module

        except KeyError:
            traceback.print_exc()
            print('problem loading custom widget')
            return None

        return getattr(self.custom_modules[file_name], widget_name)(info, *args)

    def notify_dbus(self, handler, key, value):
        proxy.updateSetting('(ssss)', self.uuid, handler.instance_id, key, json.dumps(value))

    def set_instance(self, info):
        self.instance_stack.set_visible_child_name(info["id"])
        if "stack" in info:
            self.stack_switcher.set_stack(info["stack"])
            children = info["stack"].get_children()
            if len(children) > 1:
                if self.tab in range(len(children)):
                    info["stack"].set_visible_child(children[self.tab])
                else:
                    info["stack"].set_visible_child(children[0])
        if proxy:
            proxy.highlightXlet('(ssb)', self.uuid, self.selected_instance["id"], False)
            proxy.highlightXlet('(ssb)', self.uuid, info["id"], True)
        self.selected_instance = info

    def previous_instance(self, *args):
        self.instance_stack.set_transition_type(Gtk.StackTransitionType.OVER_RIGHT)
        index = self.instance_info.index(self.selected_instance)
        self.set_instance(self.instance_info[index-1])

    def next_instance(self, *args):
        self.instance_stack.set_transition_type(Gtk.StackTransitionType.OVER_LEFT)
        index = self.instance_info.index(self.selected_instance)
        if index == len(self.instance_info) - 1:
            index = 0
        else:
            index +=1
        self.set_instance(self.instance_info[index])

    # def unpack_args(self, args):
    #    args = {}

    def backup(self, *args):
        dialog = Gtk.FileChooserDialog(_("Select or enter file to export to"),
                                       None,
                                       Gtk.FileChooserAction.SAVE,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT))
        dialog.set_do_overwrite_confirmation(True)
        filter_text = Gtk.FileFilter()
        filter_text.add_pattern("*.json")
        filter_text.set_name(_("JSON files"))
        dialog.add_filter(filter_text)

        response = dialog.run()

        if response == Gtk.ResponseType.ACCEPT:
            filename = dialog.get_filename()
            if ".json" not in filename:
                filename = filename + ".json"
            self.selected_instance["settings"].save_to_file(filename)

        dialog.destroy()

    def restore(self, *args):
        dialog = Gtk.FileChooserDialog(_("Select a JSON file to import"),
                                       None,
                                       Gtk.FileChooserAction.OPEN,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
        filter_text = Gtk.FileFilter()
        filter_text.add_pattern("*.json")
        filter_text.set_name(_("JSON files"))
        dialog.add_filter(filter_text)

        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
            self.selected_instance["settings"].load_from_file(filename)

        dialog.destroy()

    def reset(self, *args):
        self.selected_instance["settings"].reset_to_defaults()

    def reload_xlet(self, *args):
        if proxy:
            proxy.ReloadXlet('(ss)', self.uuid, self.type.upper())

    def quit(self, *args):
        if proxy:
            proxy.highlightXlet('(ssb)', self.uuid, self.selected_instance["id"], False)

        self.window.destroy()
        Gtk.main_quit()

if __name__ == "__main__":
    setproctitle("xlet-settings")
    import signal

    parser = argparse.ArgumentParser(
        description="xlet-settings - Configuration tool for Cinnamon xlets",
    )
    parser.add_argument("type", type=str, metavar="type", choices=("applet", "desklet", "extension"), help='The type of xlet to configure - applet, extension or desklet.')
    parser.add_argument("uuid", type=str, help="The UUID of the xlet")
    parser.add_argument("-i", "--id", type=int, default=None, metavar="instance_id", help="If a UUID is provided, this is the instance id.")
    parser.add_argument("-t", "--tab", type=int, default=None, metavar="tab_number", help="Tab index to open.")

    args = parser.parse_args()

    window = MainWindow(args)
    signal.signal(signal.SIGINT, window.quit)
    Gtk.main()
